///////////////////////////////////////////
//
// file:  main.cpp
//
// Copyright 2004 John R. Ashmun
//                jrashmun@halcyon.com
//
//    This file is part of MIXAL. 
//
//    MIXAL is free software; you can redistribute it and/or modify 
//    it under the terms of the GNU General Public License as
//    published by the Free Software Foundation; either version 2
//    of the License, or (at your option) any later version. 
//
//    MIXAL is distributed in the hope that it will be useful, 
//    but WITHOUT ANY WARRANTY; without even the implied warranty of 
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
//    GNU General Public License for more details. 
//
//    You should have received a copy of the GNU General Public
//    License along with MIXAL; if not, write to the
//    Free Software Foundation, Inc.
//    59 Temple Place, Suite 330
//    Boston MA 02111-1307  USA
//
///////////////////////////////////////////


#include <cstdio>
#include <iostream>

#include "MIX.h"

int main( int argc, char* argv[] )
{
   bool               a_buffer_complete = false;
   bool               eof_detected = false;
   bool               negate = false;

   char               a_buffer[67];
   char               addr_buffer[67];
   char               buffer[82];
   char               f_buffer[67];
   char               loc_buffer[11];
   char               op_buffer[5];

   int                A = 0;
   int                C = 0;
   int                F = 0;
   int                I = 0;
   int                assembled_constant = 0;

   int                character = 0;
   int                j = 0;
   int                k = 0;
   int                len = 0;
   int                loc_counter = 0;
   int                m = 0;
   int                n = 0;

   map< string, int > opcodes;
   map< string, int > f_settings;

   init_opcodes( opcodes );
   init_f_settings( f_settings );

   while( !eof_detected )
   {
      for( j = 0; j < 82; j++ )
      {
         if( (character = getchar( )) == EOF )
         {
            eof_detected = true;
            break;
         }
         else if( character == '\n' )
            break;
         else
            buffer[j] = character;
      }

      if( !eof_detected )
      {
         A = 0;
         addr_buffer[0] = '\0';
         a_buffer[0] = '\0';
         a_buffer_complete = false;
         I = 0;
         F = 0;
         f_buffer[0] = '\0';

         buffer[j] = '\0';

         j = 0;
         k = 0;
         while( buffer[j] != ' ' )
            loc_buffer[k++] = buffer[j++];

         loc_buffer[k] = '\0';

         j = 10;
         k = 0;
         while( buffer[j] != ' ' && buffer[j] != '\0' )
         {
            op_buffer[k++] = buffer[j];
            if( buffer[j++] == '\0' )
               break;
         }

         op_buffer[k] = '\0';

         k = 0;
         if( buffer[j] == ' ' )
         {
            j++;
            while( buffer[j] != ' ' )
            {
               addr_buffer[k++] = buffer[j];
               if( buffer[j++] == '\0' )
                  break;
            }
         }

         addr_buffer[k] = '\0';

         cerr << "location " << loc_counter << endl;
         cerr << "LOC:  " << loc_buffer << endl;
         cerr << "ADDRESS:  " << addr_buffer << endl;

         if( (len = strlen( addr_buffer )) )
         {
            m = 0;
            n = 0;

            for( k = 0; k < len; k++ )
            {
               if( !a_buffer_complete )
               {
                  if( (addr_buffer[k] != ',')
                      && (addr_buffer[k] != '(')
                      && (addr_buffer[k] != ' ') )
                     a_buffer[m++] = addr_buffer[k];
                  else
                  {
                     a_buffer[m] = '\0';
                     a_buffer_complete = true;
                  }
               }

               if( addr_buffer[k] == ',' )
               {
                  I = addr_buffer[++k] - '0';
               }

               if( addr_buffer[k] == '(' )
               {
                  f_buffer[0] = addr_buffer[++k];
                  if( addr_buffer[++k] == ':' )
                  {
                     f_buffer[1] = addr_buffer[k++];
                     f_buffer[2] = addr_buffer[k];
                     f_buffer[3] = '\0';
                     break;
                  }
                  else if( addr_buffer[k] != ')' )
                  {
                     f_buffer[++n] = addr_buffer[k++];
                  }

                  if( addr_buffer[k] == ')' )
                  {
                     f_buffer[++n] = '\0';
                     break;
                  }
               }
            }
         }
         a_buffer[m] = '\0';

//cerr << "a_buffer == " << a_buffer << endl;

         if( (len = strlen( a_buffer )) )
         {
            negate = false;
            for( k = 0; k < len; k++ )
            {
               if( a_buffer[k] == '-' )
                  negate = true;
               else
                  A = 10 * A + (a_buffer[k] - '0');
            }

            if( negate )
               A = -A;
         }
         cerr << "A == " << A << endl;

         cerr << "I == " << I << endl;

//cerr << "f_buffer == " << f_buffer << endl;

         if( (len = strlen( f_buffer )) )
         {
            if( len == 1 )
               F = f_buffer[0] - '0';
            else if( len == 3 )
               F = 8 * (f_buffer[0] - '0') + (f_buffer[2] - '0');
            else
            {
               F = f_buffer[0] - '0';
               for( n = 1; n < len; n++ )
                  F = 10 * F + (f_buffer[n] - '0');
            }
         }
         else
            F = f_settings[op_buffer];
        
         cerr << "F == " << F << endl;

         if( !strcmp( op_buffer, "ORIG" ) )
         {
            if( A >= 0 )
               loc_counter = A - 1;
            else
               loc_counter = A + 1;
         }
         else if( !strcmp( op_buffer, "ALF" ) )
         {
            assembled_constant = ascii_to_MIX( buffer[16] ) << 24;
            assembled_constant |= (ascii_to_MIX( buffer[17] ) << 18);
            assembled_constant |= (ascii_to_MIX( buffer[18] ) << 12);
            assembled_constant |= (ascii_to_MIX( buffer[19] ) << 6);
            assembled_constant |= ascii_to_MIX( buffer[20] );
            
            cout << loc_counter
                 << " "
                 << assembled_constant
                 << endl;
         }
         else if( !strcmp( op_buffer, "CON" ) )
         {
            assembled_constant = 0;
            n = 0;
            negate = false;
            while( addr_buffer[n] == '-' || isdigit( addr_buffer[n] ) )
            {
               if( addr_buffer[n] == '-' )
                  negate = true;
               else
               {
                  assembled_constant = 10 * assembled_constant
                                       + (addr_buffer[n] - '0');
               }
               n++;
            }

            if( negate )
            {
               cout << loc_counter
                    << " -"
                    << assembled_constant
                    << endl;
            }
            else
            {
               cout << loc_counter
                    << " "
                    << assembled_constant
                    << endl;
            }
         }
         else
         {
            C = opcodes[op_buffer];
            cerr << "OP:  " << op_buffer << "; C == " << C << endl;

            if( A < 0 )
            {
               cout << loc_counter
                    << " -"
                    << ((-A) << 18) + (I << 12) + (F << 6) + C
                    << endl;
            }
            else
            {
               cout << loc_counter
                    << " "
                    << (A << 18) + (I << 12) + (F << 6) + C
                    << endl;
            }
         }

         cerr << endl;

         if( loc_counter >= -1 )
            loc_counter++;
         else
            loc_counter--;
      }
   }

   return 0;
}
